#include "config.h"
//-----------------

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <list>

#include <sys/file.h>    // flock()

#include "ftp.h"

#include "Var.h"
#include "calcids.h"
#include "ti89util.h"

//----- the prototypes -----

void ftpCDUP(const char *args);
void ftpCWD(const char *args);
void ftpLIST(const char *args);
void ftpNLST(const char *args);
void ftp_LIST_or_NLST(bool longlist, const char *args);
void ftpPWD(const char *args);
void ftpPORT(const char *args);
void ftpPASV(const char *args);
void ftpRETR(const char *args);
void ftpSITE(const char *args);
void ftpSTOR(const char *args);
void ftpSYST(const char *args);
void ftpTYPE(const char *args);

void send_PIC_as_ASCII(int sock, Var &v);
void send_TEXT_as_ASCII(int sock, Var &v);
bool split_path(const char *p, char *folder, char *var, bool pref);
bool validate_path(const char *folder, const char *var);
void list_contents(int sock, list<Var> &fold);
void nlst_contents(int sock, list<Var> &fold);
int open_data_socket(void);

//----- the constants -----

const bool PREF_VAR = true;   // disambiguation for split_path
const bool PREF_FOLD = false;

//----- the variables -----

char Currfold[VARNAMELEN+2] = "/";  // current folder ("/" is toplevel)

int Listener;                 // listener socket for passive connections
struct sockaddr_in ListAddr;  // listener address to bind to socket
struct sockaddr_in RemData;   // address of remote data port

char Xfertype = 'A';     // data transfer type ('A' or 'I')
bool Passive = false;

//----- the list -----

struct FtpCmd {
  char *name;
  void (*func)(const char *args);
  bool nonAnon;
};

const FtpCmd Commands[] = {
  // name   func     nonAnon
  { "CDUP", ftpCDUP, 0},
  { "CWD",  ftpCWD,  0},
  { "LIST", ftpLIST, 0},
  { "NLST", ftpNLST, 0},
  { "PWD",  ftpPWD,  0},
  { "PORT", ftpPORT, 0},
  { "PASV", ftpPASV, 0},
  { "RETR", ftpRETR, 1},
  { "SITE", ftpSITE, 0},
  { "STOR", ftpSTOR, 1},
  { "SYST", ftpSYST, 0},
  { "TYPE", ftpTYPE, 0}
};

//----------------------------------------------------------------------

// FTP interface.  Return true=command found, false=command not found
bool exec_ftp_command(const char *cmd, const char *args, const bool anon)
{
    for (unsigned i = 0; i < sizeof(Commands)/sizeof(FtpCmd); i++)
    {
        if (strcmp(cmd, Commands[i].name) == 0)
	{
	    if (Commands[i].nonAnon && anon)
	        printf("550 Look; don't touch.\n");
	    else
	        Commands[i].func(args);
	    return true;
	}
    }
    return false;
}

//----------------------------------------------------------------------

void ftpCDUP(const char *args)
{
    Currfold[1] = 0;  // one level -- this is an easy one
    printf("257 \"%s\"\n", Currfold);
}

//----------------------------------------------------------------------

void ftpCWD(const char *args)
{
    char fname[VARNAMELEN+1];  // folder name
    char vname[VARNAMELEN+1];  // var name
    int retval;
  
    if (!split_path(args, fname, vname, PREF_FOLD))
    {
        printf("501 Syntax error.\n");
	return;
    }

    if (!validate_path(fname, vname))
    {
        printf("553 Name not allowed.\n");
	return;
    }

    if (vname[0] != 0) // why give a file name?
    {
        printf("550 No such folder.\n");
	return;
    }

    // no need to change if folder is current
    if (strcmp(fname,Currfold) == 0)
    {
	printf("257 \"%s\"\n", Currfold);
	return;
    }

    // "/" and "/main" always exist -- don't check
    if (fname[1] == 0 || strcmp(fname,"/main") == 0)
    {
        strcpy(Currfold, fname);
	printf("257 \"%s\"\n", Currfold);
	return;
    }

    if (flock(fileno(ti_device), LOCK_EX|LOCK_NB) == -1)
    {
        printf("450 /dev/ti is locked.\n");
	return;
    }

    retval = verifyFold(fname+1);

    if (flock(fileno(ti_device), LOCK_UN|LOCK_NB) == -1)
    {
        printf("421 Error unlocking /dev/ti.\n");
	exit(1);
    }

    if (retval < 0)
    {
        printf("451 Receive error (%d).\n", retval);
	return;
    }

    if (retval == 0)
        printf("550 No such folder.\n");
    else
    {
        strcpy(Currfold, fname);
	printf("257 \"%s\"\n", Currfold);
    }

}

//----------------------------------------------------------------------

void ftpLIST(const char *args)
{
    ftp_LIST_or_NLST(true, args);
}

//----------------------------------------------------------------------

void ftpNLST(const char *args)
{
    ftp_LIST_or_NLST(false, args);
}

//----------------------------------------------------------------------

void ftp_LIST_or_NLST(bool longlist, const char *args)
{
    char fname[VARNAMELEN+1];  // folder name
    char vname[VARNAMELEN+1];  // var name
    int sock;
    int retval;
    list <Var> varlist;

    if (!split_path(args, fname, vname, PREF_FOLD))
    {
        printf("501 Syntax error.\n");
	return;
    }

    if (!validate_path(fname, vname))
    {
        printf("553 Name not allowed.\n");
	return;
    }

    if (flock(fileno(ti_device), LOCK_EX|LOCK_NB) == -1)
    {
        printf("450 /dev/ti is locked.\n");
	return;
    }

    retval = getContents(fname+1, varlist);     // get vars on calc

    if (flock(fileno(ti_device), LOCK_UN|LOCK_NB) == -1)
    {
        printf("421 Error unlocking /dev/ti.\n");
	exit(1);
    }
	
    if (retval == -1)
    {
        printf("550 No such folder.\n", retval);
	return;
    }
    else if (retval < 0)
    {
        printf("451 Receive error (%d).\n", retval);
	return;
    }

    if (fname[1] == 0) // add special files to "/"
    {
      //varlist.push_back(Var("backup", T_BACK, 256*1024,    0));
      //varlist.push_back(Var("code",   T_CODE, 2*1024*1024, 0));
	varlist.push_back(Var("cert",   T_CERT, 267,         1));
	varlist.push_back(Var("screen", T_PIC,  240*128/8,   1));
    }

    /* create a socket for the data connection */
    sock = open_data_socket();
    if (sock == -1)
        return;

    printf("150 \"%s\" contents:\n", fname);
    fflush(stdout);
    if (longlist)
        list_contents(sock, varlist);
    else
        nlst_contents(sock, varlist);
    shutdown(sock,2);
    close(sock);
    printf("226 Transfer complete.\n");
    return;
}

//----------------------------------------------------------------------

void ftpPORT(const char *args)
{
    unsigned h1,h2,h3,h4,p1,p2;

    if (sscanf(args," %u , %u , %u , %u , %u , %u ",
	       &h1, &h2, &h3, &h4, &p1, &p2) < 6)
    {
        printf("501 Syntax error.\n");
	return;
    }

    if (h1 > 255 || h2 > 255 || h3 > 255 || h4 > 255
	|| p1 > 255 || p2 > 255)
    {
        printf("501 Range error.\n");
	return;
    }

    RemData.sin_addr.s_addr = htonl((h1<<24)+(h2<<16)+(h3<<8)+h4);
    RemData.sin_port = htons((p1<<8)+p2);
    printf("200 PORT %s:%d\n", inet_ntoa(RemData.sin_addr),
	                       ntohs(RemData.sin_port));
}

//----------------------------------------------------------------------

void ftpPASV(const char *args)
{
    unsigned addr;
    unsigned short port;

    if (Listener == -1)
    {
        printf("502 Passive mode not available.\n");
	return;
    }

    addr = ntohl(ListAddr.sin_addr.s_addr);
    port = ntohs(ListAddr.sin_port);

    printf("227 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\n", (addr>>24)&0xFF,
	   (addr>>16)&0xFF, (addr>>8)&0xFF, addr&0xFF,
	   (port>>8)&0xFF, port&0xFF);
    Passive = true;
}

//----------------------------------------------------------------------

void ftpPWD(const char *args)
{
    printf("257 \"%s\"\n", Currfold);
}

//----------------------------------------------------------------------

void ftpRETR(const char *args)
{
    char fname[VARNAMELEN+1];  // folder name
    char vname[VARNAMELEN+1];  // var name
    int sock;
    int retval;

    if (!split_path(args, fname, vname, PREF_VAR))
    {
        printf("501 Syntax error.\n");
	return;
    }

    if (!validate_path(fname, vname))
    {
        printf("553 Name not allowed.\n");
	return;
    }

    if (vname[0] == 0)
    {
        printf("550 No such file.\n");
	return;
    }

    if (flock(fileno(ti_device), LOCK_EX|LOCK_NB) == -1)
    {
        printf("450 /dev/ti is locked.\n");
	return;
    }

    Var v;              // create a variable structure

    if (fname[1] == 0) // special files in "/"
    {
        if (strcmp(vname,"screen") == 0)     // screenshot
	    retval = getScreen(v);   
	else if (strcmp(vname,"cert") == 0)  // certificate?
	    retval = getCert(v);
	else
	{
	    printf("550 No such file.\n");
	    return;
	}
    }
    else
        retval = getVar(fname+1, vname, v);

    if (flock(fileno(ti_device), LOCK_UN|LOCK_NB) == -1)
    {
        printf("421 Error unlocking /dev/ti.\n");
	exit(1);
    }

    if (retval < 0)
    {
	printf("451 Receive error (%d).\n", retval);
	return;
    }

    sock = open_data_socket();
    if (sock == -1)
        return;

    // some files may be retrieved in ASCII mode
    if (Xfertype == 'A')
    {
        if (v.type==T_PIC)
 	  send_PIC_as_ASCII(sock, v);
	else if (v.type==T_TEXT)
	  send_TEXT_as_ASCII(sock, v);
	else
	{
	  printf("550 ASCII mode retrieve not allowed for that file.\n");
	  return;
	}
    }

    else if (v.type==T_CERT)
    {
        printf("150 Sending \"%s\" (%d bytes).\n", "cert", v.length);
	fflush(stdout);
	send(sock, v.data, v.length, 0);
	shutdown(sock, 2);
	close(sock);
	printf("226 Transfer complete.\n");
    }
    else
    {
        printf("150 Sending \"%s\" (%d bytes).\n", v.name, v.length+0x58);
        fflush(stdout);

	/* send the header */
	char buf[0x56];
	memset(buf, 0, 0x56);
	if (v.calcid == ID89)
	  strcpy(buf, "**TI89**");
	else if (v.calcid == ID92)
	  strcpy(buf, "**TI92**");
	strncpy(buf+0x0A, fname+1, VARNAMELEN);
	strncpy(buf+0x40, v.name, VARNAMELEN);
	buf[0x48] = v.type;
	buf[0x4C] = (v.length + 0x58) & 0xFF;
	buf[0x4D] = ((v.length + 0x58)>>8) & 0xff;
	buf[0x4E] = ((v.length + 0x58)>>16) & 0xFF;
	buf[0x4F] = ((v.length + 0x58)>>24) & 0xFF;
	buf[0x08] = 0x01;  // magic numbers
	buf[0x3A] = 0x01;
	buf[0x3C] = 0x52;
	buf[0x50] = (char)0xA5;
	buf[0x51] = 0x5A;
	send(sock, buf, 0x56, 0);
	// send the data
	send(sock, v.data, v.length, 0);
	// send the checksum
	buf[0] = v.checksum & 0xFF;
	buf[1] = (v.checksum>>8) & 0xFF;
	send(sock, buf, 2, 0);
	// finish the transfer
	shutdown(sock, 2);
	close(sock);
	printf("226 Transfer complete.\n");
    }
}

//----------------------------------------------------------------------

void ftpSITE(const char *args)
{
    unsigned char id;
    unsigned flags;
    const char *model;
    const char *task;
    const char *batt;

    if (flock(fileno(ti_device), LOCK_EX|LOCK_NB) == -1)
    {
        printf("450 /dev/ti is locked.\n");
	return;
    }

    id = detect_calc_type(ti_device, &flags);

    if (flock(fileno(ti_device), LOCK_UN|LOCK_NB) == -1)
    {
        printf("421 Error unlocking /dev/ti.\n");
	exit(1);
    }

    switch (id)
    {
    case ID89:
      model = "TI-89";
      break;
    case ID92:
      model = "TI-92";
      break;
    default:
      printf("450 No TI-89 or TI-92 detected.\n");
      return;      
    }

    if (flags & 0x0010)
      batt = "low";
    else
      batt = "good";

    if (flags & 0x0100)
      task = "application";
    else
      task = "home screen";

    printf("220- Unit: %s\n"
	   "220- Task: %s\n"
	   "220  Batt: %s\n", model, task, batt);
}

//----------------------------------------------------------------------

void ftpSTOR(const char *args)
{
    char fname[VARNAMELEN+1];  // folder name
    char vname[VARNAMELEN+1];  // var name
    unsigned char buf[0x56];
    int sock;
    int retval;
    unsigned recvtot;
    Var v;

    if (Xfertype == 'A')
    {
        printf("550 ASCII mode stores not allowed.\n");
	return;
    }

    if (!split_path(args, fname, vname, PREF_VAR))
    {
        printf("501 Syntax error.\n");
	return;
    }

    if (!validate_path(fname, vname))
    {
        printf("553 Name not allowed.\n");
	return;
    }

    // there must be a variable name
    // also, can't store files in "/" (yet)
    if (vname[0] == 0 || fname[1] == 0)
    {
        printf("553 File name not allowed.\n");
	return;
    }

    // there is no need to check that a folder exists
    // because the calc will create it if necessary

    printf("150 Receiving \"%s/%s\".\n", fname, vname);
    fflush(stdout);

    sock = open_data_socket();
    if (sock == -1)
        return;

    // receive header
    if (recv(sock, buf, 0x56, 0) != 0x56)
    {
        printf("551 Receive error (%d).\n", 1);
	return;
    }

    // file type checks
    if (memcmp(buf,"**TI89**",8) != 0 && memcmp(buf,"**TI92**",8) != 0)
    {
        printf("551 Invalid file (%d).\n", 1);
	return;
    }

    // magic number checks
    if ((memcmp(buf+8, "\x01\x00", 2) != 0)
	|| (memcmp(buf+0x3A, "\x01\x00\x52\x00\x00\x00", 6) != 0)
	|| (memcmp(buf+0x50, "\xA5\x5A\x00\x00\x00\x00", 6) != 0))
    {
        printf("551 Invalid file (%d).\n", 2);
	return;
    }

    v.type = buf[0x48];
    v.length = (unsigned)buf[0x4C]
      + ((unsigned)buf[0x4D]<<8)
      + ((unsigned)buf[0x4E]<<16)
      + ((unsigned)buf[0x4F]<<24) - 0x58;

    v.data = new unsigned char [v.length];
    if (!v.data)
    {
        printf("552 Insufficient memory.\n");
	return;
    }

    // receive data
    recvtot = 0;
    do {
        retval = recv(sock, v.data+recvtot, v.length-recvtot, 0);
	if (retval == -1)
	{
	    printf("551 Receive error (%d).\n", 2);
	    return;
	}
	recvtot += retval;
    } while (recvtot < v.length);

    // receive checksum
    if (recv(sock, buf, 2, 0) != 2)
    {
        printf("551 Receive error (%d).\n", 3);
	return;
    }
    v.checksum = buf[0] + (((unsigned)buf[1])<<8);

    shutdown(sock, 2);
    close(sock);

    // check checksum
    unsigned checksum = 0;
    for (unsigned i = 0; i < v.length; i++)
        checksum += v.data[i];
    checksum &= 0xFFFF;

#ifdef SYSLOG_DEBUG
  syslog(LOG_DEBUG, "    File checksum: stored=%04X calculated=%04X",
	 v.checksum, checksum);
#endif

    if (checksum != v.checksum)
    {
        printf("551 Invalid file (%d).\n", 3);
	return;
    }
      
    // FINALLY, try sending to the calc

    if (flock(fileno(ti_device), LOCK_EX) == -1)   // BLOCKING LOCK
    {
        printf("450 Failed to lock /dev/ti.\n");
	return;
    }

    retval = sendVar(fname+1, vname, v);

    if (flock(fileno(ti_device), LOCK_UN|LOCK_NB) == -1)
    {
        printf("421 Error unlocking /dev/ti.\n");
	exit(1);
    }

    if (retval == -1) // calc rejected
    {
        printf("552 Insufficient memory.\n");
	return;
    }
    else if (retval < 0) // misc error
    {
	printf("451 Transmit error (%d).\n", retval);
	return;
    }

    // now for the good one
    printf("226 Transfer complete.\n");
    return;
}

//----------------------------------------------------------------------

void ftpSYST(const char *args)
{
    printf("215 UNIX\n");
}

//----------------------------------------------------------------------

void ftpTYPE(const char *args)
{
    if (toupper(args[0]) == 'A')
    {
        Xfertype = 'A';
        printf("200 Type set to ASCII.\n");
    }
    else if (toupper(args[0]) == 'I')
    {
        Xfertype = 'I';
        printf("200 Type set to IMAGE.\n");
    }
    else
        printf("504 Types A and I are supported.\n");
}

//----------------------------------------------------------------------

void send_PIC_as_ASCII(int sock, Var &v)
{
  static const char pix4[16] = {' ','`','\'','"','.','[','\\','7',
  			',','/',']','F','_','J','L','M' };

  char buf[240/2+2];  // so far the largest calc screen is 240x128
  char upper, lower;
  unsigned x, y;

  unsigned height = ((unsigned)v.data[2]<<8) + v.data[3];
  unsigned width = ((unsigned)v.data[4]<<8) + v.data[5];
  unsigned bytewidth = (width+7)/8;

  buf[(width+1)/2] = '\r';    // set up CRLF line end
  buf[(width+1)/2+1] = '\n';

  printf("150 Sending \"%s\" as %dx%d ASCII art (%d bytes).\n", v.name,
	 (width+1)/2, (height+1)/2, ((width+1)/2+2)*((height+1)/2));
  fflush(stdout);

  // do as many double-rows as possible
  for (y = 0; y < (height&~1); y += 2)
  {
    for (x = 0; x < width; x += 2)
    {
        upper = (v.data[6+y*bytewidth+x/8] >> (6-x%8)) & 3;
        lower = (v.data[6+(y+1)*bytewidth+x/8] >> (6-x%8)) & 3;
	buf[x/2] = pix4[(lower<<2)|upper];
    }
    send(sock, buf, (width+1)/2+2, 0);
  }

  // do a final row if necessary
  if (y < height)
  {
    for (x = 0; x < width; x += 2)
    {
        upper = (v.data[6+y*bytewidth+x/8] >> (6-x%8)) & 3;
	buf[x/2] = pix4[upper];
    }
    send(sock, buf, (width+1)/2+2, 0);
  }

  // finish the transfer
  shutdown(sock, 2);
  close(sock);
  printf("226 Transfer complete.\n");
}

//----------------------------------------------------------------------

void send_TEXT_as_ASCII(int sock, Var &v)
{
    char buf[80];
    unsigned pos;

    printf("150 Sending \"%s\" as text (approximately %d bytes).\n", v.name, v.length-6);
    fflush(stdout);

    pos = 0;
    for (unsigned i = 4; i < v.length-2; i++)
    {
        buf[pos++] = v.data[i];
	if (buf[pos-1] == '\r')       // replace CR by CRLF
	    buf[pos++] = '\n';
	if (pos > 78)
	{
	    send(sock, buf, pos, 0);
	    pos = 0;
	}
    }
    if (pos > 0)
        send(sock, buf, pos, 0);
    if (buf[pos-1] != '\n')
        send(sock, "\r\n", 2, 0);

    // finish the transfer
    shutdown(sock, 2);
    close(sock);
    printf("226 Transfer complete.\n");
}

//----------------------------------------------------------------------

bool split_path(const char *path, char *folder, char *var, bool pref)
{ 
    unsigned pos, fpos, i;
    unsigned ambig = 2;
    // ambig enumerates the type of ambiguity
    // 0: "/blah/foo", "blah/foo", "/blah", "blah/" (not ambiguous)
    // 1: "/blah" (one-slash ambiguity)
    // 2: "blah" (two-slash ambiguity)

    pos = 0;
    fpos = 0;
    if (path[0] == '/') {  // skip leading slash
        pos++;
	folder[fpos++] = '/';
	ambig = 1;
    }

    for (i=0; path[pos]!=0 && path[pos]!='/' && i<VARNAMELEN; i++)
        folder[fpos++] = path[pos++];
    folder[fpos++] = 0;
    
    if ((path[pos] != '/') && (path[pos] != 0))     // name too long?
	return false;
  
    if (path[pos] == '/') {     // skip slash
        pos++;
	ambig = 0;
    }

    for (i=0; path[pos]!=0 && path[pos]!='/' && i<VARNAMELEN; i++, pos++)
        var[i] = path[pos];
    var[i] = 0;

    if (path[pos] != 0)     // name too long?
	return false;

    // exchange ambiguous name if necessary
    if (ambig > 0 && pref == PREF_VAR)
    {
        if (ambig == 1) // form "/blah"
	{
	    strcpy(var, folder+1);
	    folder[0] = '/';
	    folder[1] = 0;
	}
        else // if (ambig == 2) // form "blah"
	{
	    strcpy(var, folder);
	    strcpy(folder, Currfold);
	}
#ifdef SYSLOG_DEBUG
	syslog(LOG_DEBUG, "Scanned path (%s,%s).", folder+1, var);
#endif
	return true;
    }

    // use current folder as the default
    if (folder[0] == 0)
        strcpy(folder, Currfold);
    else if (folder[0] != '/')      // make sure it starts with a slash
    {
        memmove(folder+1, folder, VARNAMELEN);
	folder[0] = '/';
    }

#ifdef SYSLOG_DEBUG
    syslog(LOG_DEBUG, "Scanned path (%s,%s).", folder+1, var);
#endif
    return true;
}

//----------------------------------------------------------------------

bool validate_path(const char *folder, const char *var)
{
  static const char charmap[256] = {
    // bit 0: 1=can be any char after the first in var or folder name
    // bit 1: 1=can be first char of var name (or folder name)
    // SPECIAL CASE: "_" is allowed to start a var name in folder "main"
//0 1 2 3 4 5 6 7  8 9 A B C D E F 10 1 2 3 4 5 6 7 18 9 A B C D E F
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, /*00-1F*/
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 1,1,1,1,1,1,1,1, 1,1,0,0,0,0,0,0, /*20-3F*/
  0,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,0,0,0,0,1, /*40-5F*/
  0,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,0,0,0,0,0, /*60-7F*/
  3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,0,0,0, 0,0,0,0,0,0,0,0, /*80-9F*/
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,3,0,0, 0,0,0,0,0,0,0,0, /*A0-BF*/
  3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,0, 3,3,3,3,3,3,3,3, /*C0-DF*/
  3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,0, 3,3,3,3,3,3,3,3  /*E0-FF*/
  };

    if (folder[1] != 0)
    {
        if ((charmap[folder[1]] & 2) == 0)
	  return false;
	for (unsigned i = 2; folder[i]; i++)
	  if ((charmap[folder[1]] & 1) == 0)
	    return false;
    }

    if (var[0] != 0)
    {
        if ((charmap[var[0]] & 2) == 0)
	{
	  if (var[0] == '_' && strcmp(folder, "/main") == 0)
	    ; // vars starting with underscore are allowed in "/main"
	  else
	    return false;
	}
	for (unsigned i = 1; var[i]; i++)
	  if ((charmap[var[1]] & 1) == 0)
	    return false;
    }

    return true;
}

//----------------------------------------------------------------------

void list_contents(int sock, list<Var> &fold)
{
    char buf[80];
    unsigned len;
    static const char *typestrings[] = {
      "expr","[01]","[02]","[03]","list","[05]","mat","[07]",
      "[08]","[09]","data","text","str","gdb","fig","[0F]",
      "pic","[11]","prgm","func","mac","[15]","[16]","[17]",
      "[18]","<19>","<1A>","[1B]","[1C]","backup","[1E]","folder",
      "cert","asm","[22]","product"};

    for (list<Var>::iterator it = fold.begin(); it != fold.end(); it++)
    {
        char chd, chx, chw;

	if (it->type==T_FOLD)
  	    chd = 'd';
	else
	    chd = '-';

	if (it->perm & 1)
  	    chw = '-';
	else
	    chw = 'w';

	if (it->type==T_FOLD || it->type==T_ASM
	    || it->type==T_PRGM || it->type==T_FUNC)
	    chx = 'x';
	else
	    chx = '-';

	len = sprintf(buf, "%cr%c%c------   1 root     %-8s %8d Feb 29  2000 %s\r\n",
		      chd, chw, chx, typestrings[it->type],
		      it->length, it->name);
	send(sock, buf, len, 0);
    }
}

//----------------------------------------------------------------------
void nlst_contents(int sock, list<Var> &fold)
{
    char buf[11];

    for (list<Var>::iterator it = fold.begin(); it != fold.end(); it++)
    {
        sprintf(buf, "%-8s\r\n", it->name);
	send(sock, buf, 10, 0);
    }
}

//----------------------------------------------------------------------

int open_data_socket(void)
{
    int sock;

    if (Passive)
    {
        struct sockaddr_in rcpt_addr;

	unsigned socksize = sizeof(sockaddr_in);
        sock = accept(Listener, (struct sockaddr*)&rcpt_addr, &socksize);
	if (sock == -1)
	{
	    printf("425 Can't open data connection.\n");
	    return -1;
	}

	if (rcpt_addr.sin_addr.s_addr != RemData.sin_addr.s_addr)
	{
	    printf("425 Can't open data connection.\n");
	    shutdown(sock, 2);
	    close(sock);
	    return -1;
	}
    }
    else // not passive
    {
        sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock == -1)
	{
	    printf("425 Can't open data connection.\n");
	    return -1;
	}

	if (connect(sock, (struct sockaddr*)&RemData, sizeof(sockaddr_in)) == -1)
	{
	    printf("425 Can't open data connection.\n");
	    close(sock);
	    return -1;
	}
    }
    return sock;
}

//---------------------------------------------------------------------- 

int setup_sockets(struct sockaddr_in &local, struct sockaddr_in &remote)
{
    /* set up default data port at (control port - 1) */ 
    RemData.sin_addr.s_addr = remote.sin_addr.s_addr;
    RemData.sin_port = htons(ntohs(remote.sin_port)-1);

    /* set up listener socket for passive connection */
    Listener = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (Listener == -1) // now passive mode is not allowed
        return 0;       // but don't return error

    ListAddr.sin_addr.s_addr = local.sin_addr.s_addr;
    ListAddr.sin_port = htons(12345);

    do {
        ListAddr.sin_port = htons(ntohs(ListAddr.sin_port)+1);
    } while (bind(Listener, (struct sockaddr*)&ListAddr, sizeof(sockaddr_in)) == -1);

    if (listen(Listener, 1) == -1)
    {
        close(Listener);
	Listener = -1;      // disallow passive mode
    }
    return 0;
}

