#include "config.h"
//-----------------

#define _XOPEN_SOURCE            // for crypt()
#include <unistd.h>

#include <ctype.h>
#include <stdio.h>
#include <string.h>

#include <pwd.h>
#ifdef SHADOW_PASSWORDS
 #include <shadow.h>
#endif

#include <sys/file.h>    // flock()

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "build.h"
#include "ftp.h"
#include "ti89util.h"

//---- prototypes -----

bool validate_user(const char *name, const char *pass);

/***********************************************************************
     Functions
*/

int main(void)
{ 
    unsigned socksize;
    struct sockaddr_in local, remote;   // sockets place holders

    char cmdbuf[CMDBUFLEN];
    char *cmdargs;

    char username[USERLEN+1] = "";
    char password[PWLEN+1] = "";
    bool user_valid = false;
    bool anonymous = true;

    int i;
    const char *calcname;

#ifdef SYSLOG
    openlog("in.ftpd89", LOG_PID, LOG_DAEMON);
#endif

    // Get addresses of local and remote machines.
    // Failure probably means that stdin is not a socket.
    socksize = sizeof(local);
    if (getsockname(fileno(stdin), (struct sockaddr*)&local, &socksize) == -1)
	exit(1);
    socksize = sizeof(remote);
    if (getpeername(fileno(stdin), (struct sockaddr*)&remote, &socksize) == -1)
	exit(1);
    // Setup stuff needed for communication functions in ftp.cpp.
    if (setup_sockets(local, remote) == -1)
        exit(1);

    // open /dev/ti and make sure there's an 89 or 92 there
    if ((ti_device = fopen("/dev/ti", "r+")) == NULL)
    {
        printf("421 Cannot open /dev/ti.\n");
        exit(1); 
    }

    if (flock(fileno(ti_device), LOCK_EX|LOCK_NB) == -1)
    {
        printf("421 /dev/ti is locked.\n");
        exit(1); 
    }
    unsigned char calcid = detect_calc_type(ti_device, 0); // don't get flags
    if (calcid == ID89)
        calcname = "TI-89";
    else if (calcid == ID92)
        calcname = "TI-92";
    else
    {
        printf("421 No TI-92 or TI-89 detected on /dev/ti.\n");
	exit(1);
    }
    sendkey(calcid, KEY_2ND | KEY_ON);  // turn off
    if (flock(fileno(ti_device), LOCK_UN|LOCK_NB) == -1)
    {
        printf("421 Error unlocking /dev/ti.\n");
        exit(1); 
    }

    printf("220 %s FTP Proxy (built "BUILDDATE")\n", calcname);

    while (1)
    {
        // flush anything that may have been thrown into
        // stdout the last time through the loop
        fflush(stdout);

        // read command line
	fgets(cmdbuf, CMDBUFLEN, stdin);

	if (strlen(cmdbuf) < 2)  // discard spurious characters
	    continue;

	// make sure the input ends in CRLF (end-of-command)
	if (cmdbuf[strlen(cmdbuf)-2] != '\r'
	    || cmdbuf[strlen(cmdbuf)-1] != '\n')
	{
	    // this should probably skip the rest of the input somehow
	    printf("500 Command too long.\n");
	    continue;
	}
	cmdbuf[strlen(cmdbuf)-2] = 0;  // remove CRLF
	
	// capitalize the command and null-terminate it
	for (i = 0; isalpha(cmdbuf[i]) ; i++)
	    cmdbuf[i] = toupper(cmdbuf[i]);
	cmdbuf[i++] = 0;

	// skip space between command and args
	for ( ; isspace(cmdbuf[i]); i++)
	  ;
	cmdargs = cmdbuf + i;

	// attempt to match command name
	if (strcmp(cmdbuf,"USER") == 0)
	{
	    if (user_valid)
	    {
	        printf("530 You are already logged in.\n");
		continue;
	    }

	    /* copy user name into buffer and issue a response*/
	    strncpy(username, cmdargs, USERLEN);
	    password[0] = 0; /* null-out password */
	    printf("331 Password required for %s.\n", username);
	}
	else if (strcmp(cmdbuf,"PASS") == 0)
	{
	    if (user_valid)
	    {
	        printf("530 You are already logged in.\n");
		continue;
	    }

	    if (username[0] == 0)
	    {
		printf("503 Login with USER first.\n");
		continue;
	    }

	    /* copy password into buffer, check it, and respond */
	    strncpy(password, cmdargs, PWLEN);
		
	    user_valid = validate_user(username, password);
	    if (user_valid)
	    {
#ifdef SYSLOG
		syslog(LOG_INFO, "User %s logged in.", username);
#endif
	        if (strcmp(username,"anonymous") != 0
		    && strcmp(username,"ftp") != 0)
		    anonymous = false;
		printf("230 User %s logged in.\n", username);
	    }
	    else
	    {
#ifdef SYSLOG
		syslog(LOG_INFO, "Failed login by %s.", username);
#endif
	        username[0] = 0;
		password[0] = 0;
		printf("530 Login incorrect.\n");
	    }
	}
	else if (strcmp(cmdbuf,"NOOP") == 0)
	    printf("200 OK\n");
	else if (strcmp(cmdbuf,"QUIT") == 0)
	{
#ifdef SYSLOG
	    if (user_valid)
	      syslog(LOG_INFO, "User %s logged out.", username);
#endif
	    printf("221 Goodbye.\n");
	    exit(0);
	}
	else
	{
	    if (!user_valid)
	      printf("530 Not logged in.\n");
	    else
	    {
#ifdef SYSLOG_DEBUG
	        syslog(LOG_DEBUG, "%s: %s %s", username, cmdbuf, cmdargs);
#endif
	        if (!exec_ftp_command(cmdbuf, cmdargs, anonymous))
		    printf("500 \"%s\": command not understood.\n", cmdbuf);
	    }
	}
    }
	
    return 0;
}

/*----------------------------------------------------------------------
*/

#ifdef SHADOW_PASSWORDS
bool validate_user(const char *name, const char *pass)
{
    struct spwd *psp;
    struct passwd *ppw;

    // there's no reason for root to use this ftp proxy
    if (strcmp(name, "root") == 0)
      return false;

    if (strcmp(name, "anonymous") == 0 || strcmp(name, "ftp") == 0)
    {
        ppw = getpwnam("ftp");
	if (ppw == NULL)
	  return false;
	if (setgid(ppw->pw_gid) == -1)
	  return false;
	if (setuid(ppw->pw_uid) == -1)
	  return false;
	return true;
    }

    setspent();
    while ((psp = (struct spwd*)getspent()) != NULL)
    {
	if (strcmp(psp->sp_namp, name) == 0)
	{
	    fflush(stdout);
	    if (strcmp(psp->sp_pwdp, (char*)crypt(pass,psp->sp_pwdp)) == 0)
	    {
		endspent();
		ppw = getpwnam(name);
		if (ppw == NULL)
		  return false;
		if (setgid(ppw->pw_gid) == -1)
		  return false;
		if (setuid(ppw->pw_uid) == -1)
		  return false;
		return true;
	    }
	    else
	    {
		endspent();
		return false;
	    }
	}
    }
    endspent();
    return false;
}
#else
char validate_user(const char *name, const char *pass)
{
    struct passwd *ppw;

    // there's no reason for root to use this ftp proxy
    if (strcmp(name, "root") == 0)
      return false;

    if (strcmp(name, "anonymous") == 0 || strcmp(name, "ftp") == 0)
    {
        ppw = getpwnam("ftp");
	if (ppw == NULL)
	    return false;
	if (setgid(ppw->pw_gid) == -1)
	    return false;
	if (setuid(ppw->pw_uid) == -1)
	    return false;
	return true;
    }

    setpwent();
    while ((ppw = (struct passwd*)getpwent()) != NULL)
    {
	if (strcmp(ppw->pw_name, name) == 0)
	{
	    if (strcmp(ppw->pw_passwd, (char*)crypt(pass,ppw->pw_passwd)) == 0)
	    {
		if (setgid(ppw->pw_gid) == -1)
		{
		    endpwent();
		    return false;
		}
		if (setuid(ppw->pw_uid) == -1)
		{
		    endpwent();
		    return false;
		}
		endpwent();
		return true;
	    }
	    else
	    {
		endpwent();
		return false;
	    }
	}
    }
    endpwent();
    return false;
}
#endif /* SHADOW_PASSWORDS */
