#include "config.h"
//-----------------

#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>

#include <time.h>
#include <list>

#include "ti89util.h"

#include "keysyms.h"
#include "calcids.h"
#include "Packet.h"
#include "Var.h"

FILE *ti_device;

//----- prototypes -----

int getVarByType(const char *folder, const char *varname, Var &var, const char type);

//-----------------------------------------------------------------------------

bool is83(FILE *dev)
{
  // Ask to receive a no-name, no-size variable.
  // The calc should respond with ACK, ABORT(EXIT).
  static unsigned char null_header[] = { 0,0,0,0,0,0 };
  Packet packet(COMP_ID(ID83),PKT_REQ_HEADER,sizeof(null_header),null_header);

  // don't use "send()" because it eats the ACK
  if (packet._send(dev) != Packet::E_OK)
    return false;

  // don't use "get()" because it sends an ACK
  if (packet._get(dev) != Packet::E_OK)
    return false;

  if (packet.type!=MSG_ACK || packet.id!=ID83)
    return false;

  if (get(packet, dev) != Packet::E_OK)
    return false;
  
  if (packet.type==PKT_ABORT && packet.length==1 && packet[0]==DATA_EXIT)
    return true;

  return false;

}

//-----------------------------------------------------------------------------

unsigned char detect_calc_type(FILE* dev, unsigned *flags)
{
  // an 86 will answer a ping from any id
  // a 92 will only answer a ping from ID92
  // an 89/92+ will answer a ping from ID92 or ID89
  Packet packet(COMP_ID(ID92), MSG_PING);

  // don't use "send()" because it eats the ACK
  if (packet._send(dev) != Packet::E_OK)
    return 0;

  // don't use "get()" because it sends an ACK
  if (packet._get(dev) != Packet::E_OK)
    return 0;

  if (packet.type != MSG_ACK)
    return 0;

  if (packet.id == ID86)
    return ID86;

  if (packet.id != ID89 && packet.id != ID92)
    return 0;

  if (flags)
    *flags = packet.length & 0xEFFF;

  if (packet.length & 0x1000)
    return ID89;
  else
    return ID92;
}

//-----------------------------------------------------------------------------

Packet::PacketError sendkey(unsigned char calcid, keysym k)
{
  // sending null locks the calc's input (both keypad & link!)
  if ( k & 0xFFFF == 0)
    return Packet::E_OK;

  return send(Packet(COMP_ID(calcid),MSG_KEY, k & 0xFFFF), ti_device);
}

//-----------------------------------------------------------------------------
// returns 1 = folder exists
//         0 = folder does not exist
//        <0 = transmission error
int verifyFold(const char *folder)
{
  static char header[] = { 0,0,0,0, 0x1A, 0 };

  Packet packet;
  unsigned char calcid;
  unsigned char compid;
  int found;
  unsigned flags;

  // ping once or twice to make sure the calc is alive
  if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
    if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
      return -1;

  if (flags != 0)
      return -2;

  compid = COMP_ID(calcid);

  send(Packet(compid,PKT_REQ_HEADER, sizeof(header), header), ti_device);

  if (get(packet, ti_device) != Packet::E_OK)  // get VAR_HEADER
    return -3;
  if (packet.type != PKT_HEADER)
    return -4;

  /*
    format of folder header:
    xx xx xx 0E 1A 04 "main" 03
    \______/ \/ \/ \/ \____/ \/
        \     \  \  \    \    \_ record terminator
         \     \  \  \    \_____ name of main folder
	  \     \  \  \_________ length of name
	   \	 \  \___________ request type
            \     \_____________ record length?
	     \__________________ data length (=packet length-4)
  */

  send(Packet(compid,MSG_CTS), ti_device);
  if (get(packet, ti_device) != Packet::E_OK)   // get DATA
    return -5;
  if (packet.type != PKT_DATA)
    return -6;

  found = 0;
  for (unsigned i = 4; i < packet.length; i += 0x0E) {
    if (strncmp(folder, (char*)&packet[i], 8) == 0) {
      found = 1;
      break;
    }
  }

  if (get(packet, ti_device) != Packet::E_OK)
    return -7;
  if (packet.type != MSG_EOT)
    return -8;

  sendkey(calcid, KEY_2ND | KEY_ON);  // turn off
  return found;
}

//-----------------------------------------------------------------------------
// returns 0 = success
//        -1 = no such folder
//       <-1 = other error
int getContents(const char *folder, list<Var> &fold)
{
  static char header[] = { 0,0,0,0, 0x19 };

  Packet packet;
  unsigned char calcid;
  unsigned char compid;
  bool found = false; // has the folder been found?
  unsigned flags;

  // ping once or twice to make sure the calc is alive
  if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
    if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
      return -2;

  if (flags != 0)
      return -3;

  compid = COMP_ID(calcid);

  send(Packet(compid,PKT_REQ_HEADER, sizeof(header), header), ti_device);

  if (get(packet, ti_device) != Packet::E_OK)  // get VAR_HEADER
    return -4;
  if (packet.type != PKT_HEADER)
    return -5;

  /*
    format of dir header:
    0E 00 00 00 00 04 "main" 07
    \______/ \/ \/ \/ \____/ \/
          \   \  \  \    \    \_ size of request packet
           \   \  \  \    \_____ name of first folder (alphabetically)
	    \   \  \  \_________ length of name
	     \	 \  \___________ var type?
              \   \_____________ 0=no records
	       \________________ data length
  */

  do {
    send(Packet(compid,MSG_CTS), ti_device);
    if (packet._get(ti_device) != Packet::E_OK)   // get DATA
      return -6;
    if (packet.type != PKT_DATA)
      return -7;

    if (packet[12] == T_FOLD)
    {
        if (found) // if the folder has been found, this is the end
	{
  	    packet.Init(compid, PKT_ABORT);  // stop transmission
	    packet._send(ti_device);
	    sendkey(calcid, KEY_2ND | KEY_ON);  // turn off
	    return 0;
	}

	int comp = memcmp(folder, &packet[4], strlen(folder));
	if (comp < 0) // if there is no hope of finding folder
	{
	    packet.Init(compid, PKT_ABORT);  // stop transmission
	    packet._send(ti_device);
	    sendkey(calcid, KEY_2ND | KEY_ON);  // turn off
	    return -1;
	}
	else if (comp == 0)
	{
	    // if the root folder is not requested, start copying var names
	    if (folder[0] != 0)
	      found = true;
	    else // if the root folder _is_ requested, copy this folder name
	      fold.push_back(Var(&packet[0]));
	}
    }
    else if (found)
    {
        fold.push_back(Var(&packet[0]));
    }

    packet.Init(compid, MSG_ACK); 
    packet._send(ti_device);

    if (get(packet, ti_device) != Packet::E_OK)
      return -8;

  } while (packet.type==MSG_MORE);

  if (packet.type != MSG_EOT)
    return -9;

  sendkey(calcid, KEY_2ND | KEY_ON);  // turn off

  if (found || folder[0] == 0) // (found is not set when folder is root)
    return 0;
  else
    return -1;
}

//-----------------------------------------------------------------------------
// returns 0 on success, <0 on error
int getScreen(Var &var)
{
  Packet packet;
  unsigned char calcid;
  unsigned char compid;
  unsigned width, height;
  unsigned i;
  unsigned flags;

  // ping once or twice to make sure the calc is alive
  if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
    if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
      return -1;

  if (flags != 0)
      return -2;

  compid = COMP_ID(calcid);

  send(Packet(compid,MSG_PRTSCR), ti_device);

  // don't check E_OK because there is a checksum error.. why?
  if (get(packet, ti_device) == Packet::E_TIMEOUT)   // get DATA
    return -3;  
  if (packet.type != PKT_DATA)
    return -4;

  if (calcid == ID89)
  {
    width = 160;
    height = 100;
  }
  else if (calcid == ID92)
  {
    width = 240;
    height = 128;
  }
  else
    return -5;

  var.length = 7 + width*height/8;
  var.type   = T_PIC;
  var.calcid = calcid;
  strcpy(var.name,"screen");

  var.data = new unsigned char [var.length];    // copy and checksum
  if (var.data == 0)
    return -6;

  var.data[0] = ((var.length-2)>>8) & 0xFF;
  var.data[1] = (var.length-2) & 0xFF;
  var.data[2] = (height>>8) & 0xFF;
  var.data[3] = height & 0xFF;
  var.data[4] = (width>>8) & 0xFF;
  var.data[5] = width & 0xFF;
  i = 6;
  for (unsigned y = 0; y < height; y++)
    for (unsigned xb = 0; xb < width/8; xb++)
      var.data[i++] = packet[y*240/8+xb];
  var.data[var.length-1] = 0xDF;                // PIC marker?

  unsigned checksum = 0;
  for (unsigned i = 0; i < var.length; i++)
      checksum += var.data[i];
  var.checksum = checksum & 0xFFFF;

  sendkey(calcid, KEY_2ND | KEY_ON);  // turn off
  return 0;
}

//-----------------------------------------------------------------------------

int getCert(Var &var)
{
    // get certificate?
    return getVarByType("", "cert", var, T_CERT);
}

//-----------------------------------------------------------------------------

int getCode(Var &var)
{
    // get product code (ROM)
    return getVarByType("", "code", var, T_CODE);  // get any type
}

//-----------------------------------------------------------------------------

int getVar(const char *folder, const char *varname, Var &var)
{
    return getVarByType(folder, varname, var, 0);  // get any type
}

//-----------------------------------------------------------------------------

int getVarByType(const char *folder, const char *varname, Var &var, const char type)
{
    static char header[24] = { 0,0,0,0 };
    // the first four chars should remain zero

  Packet packet;
  unsigned char varnamelength;
  unsigned char calcid;
  unsigned char compid;
  unsigned flags;

  // ping once or twice to make sure the calc is alive
  if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
    if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
      return -1;

  if (flags != 0)
      return -2;

  compid = COMP_ID(calcid);

  // construct var request header
  header[4] = type;
  header[5] = sprintf(header+6, "%s\\%s", folder, varname);

  send(Packet(compid,PKT_REQ_HEADER, header[5]+6, header),
       ti_device);

  if (get(packet, ti_device) != Packet::E_OK)   // get VAR_HEADER
    return -3;
  if (packet.type != PKT_HEADER)
    return -4;

  /* header format
    xx xx xx xx xx xx "..." xx
    \______/ \/ \/ \/ \___/ \/
        \     \  \  \    \   \_ record terminator
         \     \  \  \    \____ name
	  \     \  \  \________ length of name (0-8)
	   \	 \  \__________ variable type
	    \     \____________ record length (including term, 0=no records)
    	     \_________________ data length (=next packet length-4)
  */

  var.calcid = calcid;
  var.length = packet[0] + (packet[1]<<8) + (packet[2]<<16);
  var.type   = packet[4];
  varnamelength = packet[5];
  for (int i = 0; i < varnamelength; i++)
    var.name[i] = packet[6+i];
  var.name[varnamelength] = 0;

#ifdef SYSLOG_DEBUG
  syslog(LOG_DEBUG, "    Var: %s len=%04X type=%02X",
	 var.name, var.length, var.type);
#endif

  send(Packet(compid,MSG_CTS), ti_device);
  if (get(packet, ti_device) != Packet::E_OK)   // get VAR_DATA
    return -5;
  if (packet.type != PKT_DATA)
    return -6;

  var.data = new unsigned char [var.length];    // copy and checksum
  if (var.data == 0)
    return -7;
  unsigned checksum = 0;
  for (unsigned i = 0; i < var.length; i++)
  {
      var.data[i] = packet[i+4];
      checksum += var.data[i];
  }
  var.checksum = checksum & 0xFFFF;

  if (get(packet, ti_device) != Packet::E_OK)   // get EOT
  {
    delete [] var.data;
    var.data = 0;
    return -8;
  }

  sendkey(calcid, KEY_2ND | KEY_ON);  // turn off
  return 0;
}

//-----------------------------------------------------------------------------
// return 0 = OK
//       -1 = variable rejected (not enough memory)
//      <-1 = other
int sendVar(const char *folder, const char *varname, Var &v)
{
  static char header[24] = { 0,0,0,0 };

  Packet packet;
  unsigned char calcid;
  unsigned char compid;
  unsigned flags;

  // ping once or twice to make sure the calc is alive
  if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
    if ((calcid = detect_calc_type(ti_device, &flags)) == 0)
      return -2;

  if (flags != 0)
     return -3;

  compid = COMP_ID(calcid);

  header[0] = v.length & 0xFF;
  header[1] = (v.length>>8) & 0xFF;
  header[2] = (v.length>>16) & 0xFF;
  header[4] = v.type;
  header[5] = sprintf(header+6, "%s\\%s", folder, varname);

  char *data = new char [v.length+4];
  if (!data)
    return -4;
  data[0] = 0;
  data[1] = 0;
  data[2] = 0;
  data[3] = 0;
  memcpy(data+4, v.data, v.length);

#ifdef SYSLOG_DEBUG
  syslog(LOG_DEBUG, "    Var: %s len=%04X type=%02X",
	 header+6, v.length, v.type);
#endif

  // it's 7+header[5] (not 6+) because of null termination on the name
  if (send(Packet(compid,PKT_SEND_HEADER, 7+header[5], header), ti_device)
      == Packet::E_REJ)
  {
      sendkey(calcid, KEY_2ND | KEY_ON);  // turn off
      return -1;
  }

  // get MSG.CTS (hopefully)
  if (get(packet, ti_device) != Packet::E_OK) // not enough memory
    return -5;
  if (packet.type != MSG_CTS)
    return -6;

  if (send(Packet(compid,PKT_DATA, v.length+4, data),
	   ti_device) != Packet::E_OK)
    return -7;
 
  packet.Init(compid,MSG_EOT);  // use _send because no ACK is sent
  packet._send(ti_device);

  sendkey(calcid, KEY_2ND | KEY_ON);  // turn off
  return 0;
}
